/*
 * sam9g20_tlv320aic23  --  SoC audio for AT91SAM9G20-based
 * 			ATMEL AT91SAM9G20ek board.
 *
 *  Copyright (C) 2005 SAN People
 *  Copyright (C) 2008 Atmel
 *
 * Authors: Sedji Gaouaou <sedji.gaouaou@atmel.com>
 *
 * Based on ati_b1_tlv320aic23.c by:
 * Frank Mandarino <fmandarino@endrelia.com>
 * Copyright 2006 Endrelia Technologies Inc.
 * Based on corgi.c by:
 * Copyright 2005 Wolfson Microelectronics PLC.
 * Copyright 2005 Openedhand Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/kernel.h>
#include <linux/clk.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/platform_device.h>

#include <linux/atmel-ssc.h>

#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/pcm_params.h>
#include <sound/soc.h>
#include <sound/soc-dapm.h>
#include <linux/clk.h>

#include <mach/hardware.h>
#include <mach/gpio.h>
#include <mach/at91sam9_smc.h>
#include <mach/board.h>

#include "../codecs/tlv320aic23.h"
#include "atmel-pcm.h"
#include "atmel_ssc_dai.h"

#define CODEC_CLOCK 	12000000

static int m518_startup(struct snd_pcm_substream *substream)
{
	struct snd_soc_pcm_runtime *rtd = snd_pcm_substream_chip(substream);
	struct snd_soc_dai *codec_dai = rtd->dai->codec_dai;
	int ret;

	/* codec system clock is supplied by PCK0, set to 12MHz */
	ret = snd_soc_dai_set_sysclk(codec_dai, 0 ,
		CODEC_CLOCK, SND_SOC_CLOCK_IN);
	if (ret < 0)
		return ret;

	return 0;
}

static void m518_shutdown(struct snd_pcm_substream *substream)
{
	struct snd_soc_pcm_runtime *rtd = snd_pcm_substream_chip(substream);

	dev_dbg(rtd->socdev->dev, "shutdown");
}

static int m518_hw_params(struct snd_pcm_substream *substream,
	struct snd_pcm_hw_params *params)
{
	struct snd_soc_pcm_runtime *rtd = substream->private_data;
	struct snd_soc_dai *codec_dai = rtd->dai->codec_dai;
	struct snd_soc_dai *cpu_dai = rtd->dai->cpu_dai;
	struct atmel_ssc_info *ssc_p = cpu_dai->private_data;
	struct ssc_device *ssc = ssc_p->ssc;
	int ret;

	unsigned int rate;
	int cmr_div, period;

	if (ssc == NULL) {
		printk(KERN_INFO "m518_hw_params: ssc is NULL!\n");
		return -EINVAL;
	}

	/* set codec DAI configuration */
	ret = snd_soc_dai_set_fmt(codec_dai, SND_SOC_DAIFMT_I2S |
		SND_SOC_DAIFMT_NB_NF | SND_SOC_DAIFMT_CBS_CFS/*S*/);
	if (ret < 0)
		return ret;

	/* set cpu DAI configuration */
	ret = snd_soc_dai_set_fmt(cpu_dai, SND_SOC_DAIFMT_I2S |
		SND_SOC_DAIFMT_NB_NF | SND_SOC_DAIFMT_CBS_CFS/*S*/);
	if (ret < 0)
		return ret;

	/*
	 * The SSC clock dividers depend on the sample rate.  The CMR.DIV
	 * field divides the system master clock MCK to drive the SSC TK
	 * signal which provides the codec BCLK.  The TCMR.PERIOD and
	 * RCMR.PERIOD fields further divide the BCLK signal to drive
	 * the SSC TF and RF signals which provide the codec DACLRC and
	 * ADCLRC clocks.
	 *
	 * The dividers were determined through trial and error, where a
	 * CMR.DIV value is chosen such that the resulting BCLK value is
	 * divisible, or almost divisible, by (2 * sample rate), and then
	 * the TCMR.PERIOD or RCMR.PERIOD is BCLK / (2 * sample rate) - 1.
	 */
	rate = params_rate(params);

	switch (rate) {
	case 32000:
		cmr_div = 66;	/* BCLK = 133MHz/(2*66) = 1.007MHz */
		period = 15;	/* LRC = BCLK/(2*(15+1)) = 31486,742Hz */
		break;
	case 48000:
		cmr_div = 33;	/* BCLK = 133MHz/(2*33) = 2.015MHz */
		period = 20;	/* LRC = BCLK/(2*(20+1)) = 47979,79Hz */
		break;
	case 44100:
		cmr_div = 29;	/* BCLK = 133MHz/(2*29) = 2.293MHz */
		period = 25;	/* LRC = BCLK/(2*(25+1)) = 44098Hz */
		break;
	default:
		return 0;
	}
	
	/* set the MCK divider for BCLK */
	ret = snd_soc_dai_set_clkdiv(cpu_dai, ATMEL_SSC_CMR_DIV, cmr_div);
	if (ret < 0) {
		
		return ret;
	}

	if (substream->stream == SNDRV_PCM_STREAM_PLAYBACK) {
		/* set the BCLK divider for DACLRC */
		ret = snd_soc_dai_set_clkdiv(cpu_dai,
						ATMEL_SSC_TCMR_PERIOD, period);
	} else {
		/* set the BCLK divider for ADCLRC */
		ret = snd_soc_dai_set_clkdiv(cpu_dai,
						ATMEL_SSC_RCMR_PERIOD, period);
	}
	if (ret < 0){
		return ret;
	}

	return 0;
}

static struct snd_soc_ops m518_ops = {
	.startup = m518_startup,
	.hw_params = m518_hw_params,
	.shutdown = m518_shutdown,
};


static const struct snd_soc_dapm_widget m518_dapm_widgets[] = {
	SND_SOC_DAPM_HP("Headphone Jack", NULL),
	SND_SOC_DAPM_LINE("Line In", NULL),
	SND_SOC_DAPM_MIC("Mic Jack", NULL),
};

static const struct snd_soc_dapm_route intercon[] = {
	{"Headphone Jack", NULL, "LHPOUT"},
	{"Headphone Jack", NULL, "RHPOUT"},

	{"LLINEIN", NULL, "Line In"},
	{"RLINEIN", NULL, "Line In"},

	{"MICIN", NULL, "Mic Jack"},
};

/*
 * Logic for a tlv320aic23 as connected on a m518 board.
 */
static int m518_tlv320aic23_init(struct snd_soc_codec *codec)
{
	/* Add specific widgets */
	snd_soc_dapm_new_controls(codec, m518_dapm_widgets,
				  ARRAY_SIZE(m518_dapm_widgets));
	/* Set up specific audio path interconnects */
	snd_soc_dapm_add_routes(codec, intercon, ARRAY_SIZE(intercon));

	snd_soc_dapm_enable_pin(codec, "Headphone Jack");
	snd_soc_dapm_enable_pin(codec, "Line In");
	snd_soc_dapm_enable_pin(codec, "Mic Jack");

	snd_soc_dapm_sync(codec);

	return 0;
}

static struct snd_soc_dai_link m518_dai = {
	.name = "TLV320AIC23",
	.stream_name = "AIC23",
	.cpu_dai = &atmel_ssc_dai[0],
	.codec_dai = &tlv320aic23_dai,
	.init = m518_tlv320aic23_init,
	.ops = &m518_ops,
};
#if 1
static struct snd_soc_card snd_soc_m518 = {
	.name = "M518 TLV320AIC23",
	.platform = &atmel_soc_platform,
	.dai_link = &m518_dai,
	.num_links = 1,
};
#else

/* Audio machine driver */
static struct snd_soc_machine snd_soc_machine_m605 = {
	.name = "M518 TLV320AIC23",
	.dai_link = &m518_dai,
	.num_links = 1,
};
#endif

/* tlv320 audio private data */
//static struct aic23_setup_data tlv320_aic23_setup = {
//	.i2c_address = 0x1b,
//};

static struct snd_soc_device m518_snd_devdata = {
	.card = &snd_soc_m518,
	//.machine = &snd_soc_machine_m605,
	.codec_dev = &soc_codec_dev_tlv320aic23,
	//.platform = &atmel_soc_platform,
	//.codec_data = &tlv320_aic23_setup,
};

static struct platform_device *m518_snd_device;

static int __init m518_init(void)
{
	struct atmel_ssc_info *ssc_p = m518_dai.cpu_dai->private_data;
	struct ssc_device *ssc = NULL;
	int ret;

	
//#if defined(CONFIG_SND_AT91_SOC_SAM9G20_TLV320AIC23_MODULE)
	struct clk *pck0, *pll;

	at91_set_B_periph(AT91_PIN_PC1, 0);
	pck0 = clk_get(NULL, "pck0" );
	pll = clk_get(NULL, "plla" );
	clk_set_parent(pck0, pll);
	clk_set_rate(pck0, 18432000);
	clk_enable(pck0);
	at91_set_A_periph(AT91_PIN_PB17, 1);
	at91_set_A_periph(AT91_PIN_PB16, 1);
	at91_set_A_periph(AT91_PIN_PB18, 1);
//#endif
	/*
	 * Request SSC device
	 */
	ssc = ssc_request(0);
	if (IS_ERR(ssc)) {
		ret = PTR_ERR(ssc);
		ssc = NULL;
		goto err_ssc;
	}
	ssc_p->ssc = ssc;
	#if 0
	/*
	 * Configure MCLK for WM8510
	 */
	_gclk0 = clk_get(NULL, "gclk0");
	if (IS_ERR(_gclk0)) {
		_gclk0 = NULL;
		goto err_gclk0;
	}
	_pll0 = clk_get(NULL, "pll0");
	if (IS_ERR(_pll0)) {
		_pll0 = NULL;
		goto err_pll0;
	}
	if (clk_set_parent(_gclk0, _pll0)) {
		pr_warning("snd-soc-playpaq: "
			   "Failed to set PLL0 as parent for DAC clock\n");
		goto err_set_clk;
	}
	
	clk_set_rate(CODEC_CLK, 12000000);
	clk_enable(CODEC_CLK);
#endif
	m518_snd_device = platform_device_alloc("soc-audio", -1);
	if (!m518_snd_device) {
		printk(KERN_DEBUG
				"platform device allocation failed\n");
		ret = -ENOMEM;
	}

	platform_set_drvdata(m518_snd_device,
			&m518_snd_devdata);
	m518_snd_devdata.dev = &m518_snd_device->dev;

	ret = platform_device_add(m518_snd_device);
	if (ret) {
		printk(KERN_DEBUG
				"platform device allocation failed\n");
		platform_device_put(m518_snd_device);
	}

	return ret;

err_ssc:
	return ret;
}

static void __exit m518_exit(void)
{
	struct atmel_ssc_info *ssc_p = m518_dai.cpu_dai->private_data;
	struct ssc_device *ssc;

	if (ssc_p != NULL) {
		ssc = ssc_p->ssc;
		if (ssc != NULL)
			ssc_free(ssc);
		ssc_p->ssc = NULL;
	}

	platform_device_unregister(m518_snd_device);
	m518_snd_device = NULL;
}

module_init(m518_init);
module_exit(m518_exit);

/* Module information */
MODULE_AUTHOR("Ace Yang <ace.yang@artila.com>");
MODULE_DESCRIPTION("ALSA SoC Matrix518 TLV320AIC23");
MODULE_LICENSE("GPL");
